module GraphQuiz where

import qualified Data.Set as S
import qualified Data.Map as M
import Data.Map ((!))
import Data.List
import Data.Maybe

data Possibilities = NoWay | JustOne | Exactly Int | TooMany

data Tightness = TotallyTight | MightConstrain | JustTestable

class Slot a where
	choices       :: a -> Possibilities
	splitable     :: a -> Bool
	splitPoss     :: a -> (a,a)

data Function s = Function {
	tightness  :: Tightness,
	test       :: [s] -> s -> Maybe Bool,
	run        :: [s] -> s -> Maybe s
	}

data Edge i s = Edge {
	sources    :: [i],
	target     :: i,
	func	   :: Function s
	}

testEdge :: (Slot s, Ord i) => M.Map i s -> Edge i s -> Maybe Bool
testEdge m e = test (func e) (map (m!) (sources e)) (m!target e)

runEdge :: (Slot s, Ord i) => M.Map i s -> Edge i s -> Maybe s
runEdge m e = run (func e) (map (m!) (sources e)) (m!target e)


applyEdge :: (Slot s, Ord i) => M.Map i s -> Edge i s -> Maybe (M.Map i s)
applyEdge m e = do newVal <- runEdge m e
		   return $ M.insert (target e) newVal m

splitSlot :: (Slot s, Ord i) => M.Map i s -> i -> (M.Map i s, M.Map i s)
splitSlot m i = (M.insert i s1 m, M.insert i s2 m)
  where (s1,s2) = splitPoss (m!i)

findSplittable :: (Slot s, Ord i) => M.Map i s -> Maybe i
findSplittable m = listToMaybe $ map fst splitables
  where slots = M.toList m
        splitables = filter (splitable.snd) slots

checkState :: (Slot s, Ord i) =>  M.Map i s -> [Edge i s] -> Bool
checkState m = and . catMaybes . map (testEdge m)

bruteForce :: (Slot s, Ord i) => M.Map i s -> [Edge i s] -> Maybe (M.Map i s)
bruteForce m es | checkState m es = Just m
                | otherwise       = case findSplittable m of
	                              Just i -> let (m1, m2) = splitSlot m i
				      		    try1 = bruteForce m1 es
				      		    try2 = bruteForce m2 es
						in maybe try2 Just try1
				      Nothing -> Nothing

instance Slot Int where
	choices i = JustOne
	splitable   i = False
	splitPoss   i = undefined
