#!/usr/bin/python
# -*- coding: utf-8 -*- 

# Copyright (c) 2008 Manuel Holtgrewe, Mathias Ziebarth, Lucas Lürich, Joachim Breitner
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import sevendown.pretty as sp
import sevendown.graph as sg


class NullLogger(object):
    """
    Base class for loggers, with stub implementations.
    """

    def notDoneBecauseOf(self, cell):
        """
        The code checked whether the problem is solved, and it
        was not because of the given cell
        """
        pass

    def startingSolver(self, field):
        """
        A solver was started on the given field
        """
        pass

    def solvingStep(self, field):
        """
        One solver step has passed
        """
        pass

    def applying(self, rule):
        """
        A rule is about to be applied
        """
        pass

    def ruleResult(self, rule, res, field):
        """
        A rule has returned, returning res, with field
        """
        pass

    def consideringSplits(self, cell, splits):
        """
        The splits of a cell are enumerated
        """
        pass

    def applyingSplit(self, cell, split):
        """
        A certain split is applied to a cell
        """
        pass

    def willTryRule(self,rule):
        """
        A rule has been added to the todo list
        """
        pass
    
    def contradictionOccured(self, rule, cell):
        """
        A contradiction was detected by rule in cell.
        """
        pass
    

class FileLogger(NullLogger):

    def __init__(self, file):
        self.f = file

    def notDoneBecauseOf(self, cell):
        self.f.write("Not done, because %s is not done\n" % str(cell))

    def startingSolver(self, field):
        self.f.write(sp.FieldPrinter(field).formatRules())

    def solvingStep(self, field):
        self.printField(field)

    def printField(self, field):
        self.f.write("-"*78+"\n")
        self.f.write("Current Field\n")
        self.f.write(sp.FieldPrinter(field).run())
        self.f.write("\n")
    
    def applying(self, rule):
        self.f.write("Applying rule %s...\n" % (rule))

    def ruleResult(self, rule, res, field):
        if res:
            self.f.write("Rule %s changed cells:\n" % rule)
            for cell in res:
                if isinstance(cell, sg.Number):
                    self.f.write("\t" + sp.CellPrinter(cell).run())
                else:
                    self.f.write("\t%s\n" % cell)

            self.printField(field)

    def consideringSplits(self, cell, splits):
        self.f.write("split returned %s\n" % (splits))

    def applyingSplit(self, cell, split):
        self.f.write("Splitting %s to %s\n" % (cell, split))

    def willTryRule(self, rule):
        self.f.write("Going to try this rule: %s\n" % str(rule))
    
    def contradictionOccured(self, rule, cell):
        tpl = "Contradiction occured for cell %s after applying rule %s\n"
        self.f.write(tpl % (cell, rule))
    


logger = NullLogger()

# vim:sw=4:ts=4:sts=4:expandtab
