#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2008 Manuel Holtgrewe, Mathias Ziebarth, Lucas Lürich, Joachim Breitner
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""Runs the whole program against the examples in the directory 
examples.

For each example, there must be an input file and an output file.  The input 
file will be given to the program as input and the output file's content will 
be compared to the program's output.

"""

VERBOSE = False
"""Program in and out file contents iff True."""

import time
import sys
import os
import os.path
import StringIO

from optparse import OptionParser

import sevendown.loader
import sevendown.solver
import sevendown.logger

solvers = [sevendown.solver.OrderedSolver, sevendown.solver.StackSolver]

def run(input, output, solverClass, debug, verbose = False, filename = None):
    fname = os.path.basename(filename)
    try:
        loader = sevendown.loader.Loader()
        problem = loader.load(input)

        if debug:
            sevendown.logger.logger = sevendown.logger.FileLogger(sys.stdout)
        
        solver = solverClass()
        start = time.time()
        solution = solver.solve(problem)
        end = time.time()

        if debug:
            sevendown.logger.logger = sevendown.logger.NullLogger()

        if verbose:
            print "%s took %.4fs" % (solverClass, end-start)
            print solver.statistics.toString()
        
            if not solution:
                if output != "\n":
                    print "No solution found!"
                else:
                    print "No solution found, as expected."
            else:
                formatted = solution.formatResult()
                if formatted != output:
                    print "Result: Difference!"
                    print "Was:"
                    print "***"
                    print formatted
                    print "***"
                    print "Expected:"
                    print "***"
                    print output
                    print "***"
                else:
                    print "Result: OK"
        else: # not verbose
            if solution:
                formatted = solution.formatResult()
                if formatted == output:
                    print "%12s (%.4fs; %s)" % ("OK", end-start, fname)
                else:
                    print "%12s (%.4fs; %s)" % ("WRONG", end-start, fname)
            else: # no solution
                if output != "\n":
                    print "%12s (%.4fs; %s)" % ("NO SOLUTION", end-start, fname)
                else:
                    print "%12s (%.4fs; %s)" % ("OK", end-start, fname)

    except KeyboardInterrupt:
        print "%12s (%s)" % ("ABORTED", fname)
    except sevendown.solver.SolverNotApplicable, e:
        print "Solver not applicable: %s" % str(e)
    except Exception, e:
        print "%12s (%s)" % ("EXCEPTION", fname)
        sys.excepthook(*sys.exc_info())


def run_file(in_path, debug = False, solver = None, verbose = False):
    if verbose:
        print "Running test %s..." % os.path.basename(in_path)
    
    out_path = in_path.replace(".in", ".out")
            
    if not os.path.isfile(in_path):
        print "There is no file %s!" % (in_path)
        return
            
    # Skip in file without out file.  Out files without in files are
    # ignored automatically since we only search for in files...
    if not os.path.isfile(out_path):
        print "NOTE: There is no out file %s for in file %s!" % \
            (out_path, in_path)
        return
            
    # Run tests if both in file and out file exist.
    if verbose:
        print "---------------------------------------------------"
        print "In: %s" % in_path
    in_handle = open(in_path, "r")
    in_fcontents = in_handle.read()
    in_handle.close()
    if VERBOSE:
        print "***\n%s\n***" % in_fcontents
    
    if verbose:
        print "Out: %s" % out_path
    out_handle = open(out_path, "r")
    out_fcontents = out_handle.read()
    out_handle.close()
    if VERBOSE:
        print "***\n%s\n***" % out_fcontents
    
    in_io = StringIO.StringIO(in_fcontents)
    # print >>in_io, in_fcontents

    run(in_io, out_fcontents, solver, debug, verbose = verbose, filename=in_path)

    if verbose:
        print "---------------------------------------------------"


def run_all(solver = None, verbose = False):
    path_base = os.path.join(os.getcwd(), 'tests') 
    #path_base = os.path.dirname(os.path.abspath(system_test.__path__))
    path_examples = os.path.join(path_base, 'examples')
    
    print "Running all tests within '%s'..." % path_examples
    
    for root, dirs, files in os.walk(path_examples):
        in_filenames = [f for f in files if f.endswith(".in")]
        for in_filename in in_filenames:
            in_path = os.path.join(path_examples, in_filename)
            run_file(in_path, solver = solver, verbose = verbose)


if __name__ == "__main__":
    usage = "usage: %prog [options] <test_file.in>"
    parser = OptionParser(usage=usage)
    parser.add_option("-d", "--debug", dest="debug", default=False,
                  help="Write debugging output", action="store_true")
    parser.add_option("-v", "--verbose", dest="verbose", default=False,
                  help="Be a bit more verbose)", action="store_true")
    parser.add_option("-s", "--solver", dest="solver", 
                  default="StackSolver", help="Solver class to use")

    (options, args) = parser.parse_args()

    solver = None    
    for solver_class in solvers:
        if solver_class.__name__ == options.solver:
            solver = solver_class
    assert solver, "No solver with name %s found" % options.solver

    if len(args) > 0:
        run_file(args[0], debug = options.debug, solver = solver,
            verbose = options.verbose)
    else:
        run_all(solver = solver, verbose = options.verbose)

# vim:sw=4:ts=4:sts=4:expandtab
