#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2008 Manuel Holtgrewe, Mathias Ziebarth, Lucas Lürich, Joachim Breitner
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import unittest

import test_utils
import sevendown.rules as sr
import sevendown.graph as sg
import sevendown.util as su
import sevendown.config as sc
import sevendown.undo as undo

class FieldTest(test_utils.TestCase):
    def testInit(self):
        """Test that instantiating a Field does the correct initialization."""
        field = sg.Field(10, 5)
        
        # test width and height
        self.failUnlessEqual(field.width, 5)
        self.failUnlessEqual(field.height, 10)
        
        # test field.digits
        self.failUnless(hasattr(field, 'digits'))
        
        self.failUnlessEqual(len(field.digits), 10)
        for t in field.digits:
            self.failUnlessEqual(len(t), 5)
            for d in t:
                self.failUnless(isinstance(d, sg.Digit))
                self.assertEqual(d.field, field)
        
        # makes sure that the various digits are actually
        # different object
        self.failIfEqual(field.digits[0][0], field.digits[0][1])
        self.failIfEqual(field.digits[0][0], field.digits[1][0])
        
        # test attributes cells and numbers
        self.failUnlessEqual(field.numbers, [])
        self.failUnless(isinstance(field.cells, list))
        self.failUnlessEqual(len(field.cells), 50)
    
    def testCreateNumber(self):
        """Test that creating a number works."""
        field = sg.Field(1, 5)
        
        number = field.createNumber('A_w', 1, 0, 3, True)
        self.failIfEqual(number, None)
        self.assertEqual(number.field, field)
        
        for i in range(0, 3):
            self.failUnless(number.digits[i] is field.digits[0][1 + i])

        self.assert_(number.hasDigits())
        
        self.failUnless(number in field.numbers)
        self.failUnless(number in field.cells)
    
    def testCreateIntermediateNumber(self):
        """Test that creating an IntermediateNumber works."""
        field= sg.Field(1,1)
        self.assertEqual(len(field.cells), 1) # the digit!
        
        number = field.createIntermediateNumber("X_1")
        self.assertEqual(number.field, field)
        # Uncomment this test once we use Number() here as well:
        #self.failIf(number.hasDigits())
        
        self.assertNotEqual(number, None)
        self.assertEqual(len(field.intermediate_numbers), 1)
        self.assert_(number in field.intermediate_numbers)
        self.assertEqual(len(field.cells), 2)
        self.assert_(number in field.cells)
    
    def testOutput(self):
        """Test that printing a Field gives the correct output."""
        field = sg.Field(3, 4)
        
        c = 1
        for row in field.digits:
            for d in row:
                d.setFeasibleValues(set([c]))
                c = (c+1)%10
        
        ret = field.formatResult()
        exp = "1234\n5678\n9012\n"
        
        self.failUnlessEqual(ret, exp)
    
    def testNumberForNameQueryForNumber(self):
        """Test that numberForName works correctly."""
        field = sg.Field(1, 5)
        
        number = field.createNumber('A_w', 1, 0, 3, True)
        
        self.assertEqual(field.numberForName('A_w'), number)
    
    def testNumberForNameQueryForIntermediateNumber(self):
        """Test that numberForName works correctly."""
        field = sg.Field(5, 1)
        
        number = field.createIntermediateNumber('X_0')
        
        self.assertEqual(field.numberForName('X_0'), number)
    
    def testRegisterRule(self):
        field = sg.Field(5, 1)
        rule = sr.UnaryRule(sg.Digit())
        
        field.registerRule(rule)
        
        self.assertEqual(rule.field, field)
        self.assert_(rule in field.rules)
    
    def testGetNumbers(self):
        """Test Field.getNumbers()."""
        # result should be equal to the Field's numbers but not the same
        field = sg.Field(2,1)
        n1 = field.createNumber('A_w', 0, 0, 1, True)
        n2 = field.createNumber('A_w', 0, 1, 1, True)
        
        res = field.getNumbers()
        self.assertEqual(res, [n1, n2])
        
        self.assert_(field.getNumbers() is not res)
    

class FixedTest(test_utils.TestCase):
    def setUp(self):
        test_utils.TestCase.setUp(self)
        self.f = sg.Fixed(1)
    
    def testInit(self):
        """Tests the initialization."""
        self.failUnlessEqual(self.f.value(), 1)
        self.assert_(type(self.f.getFeasibleValues), set)
    
    def testCard(self):
        """Tests the cardinality."""
        self.failUnlessEqual(self.f.cardinality(), 1)
    
    def testDone(self):
        """Tests whether isDone works."""
        self.assert_(self.f.isDone())
    
    def testIsEmpty(self):
        """Tests whether isEmpty works."""
        self.assert_(not self.f.isEmpty())
    
    def testSplit(self):
        """Tests splittable == False"""
        self.failIf(self.f.isSplittable())
    
    def testStringConversion(self):
        f1 = sg.Fixed(1)
        self.assertEqual(str(f1), '1')
        
        f2 = sg.Fixed(100)
        self.assertEqual(str(f2), '100')
    
    def testMinMax(self):
        f = sg.Fixed(100)
        
        self.assertEqual(f.minimum(), 100)
        self.assertEqual(f.maximum(), 100)
    

class DigitTest(test_utils.TestCase):
    """Test the Digit class."""
    
    def setUp(self):
        test_utils.TestCase.setUp(self)
        self.field = sg.Field(1,1)
    
    def testInitNoParam(self):
        """Tests the initialization without any parameters."""
        digit = sg.Digit()
        
        self.assertEqual(digit.getFeasibleValues(), set(range(0, 10)))
        self.assertEqual(digit.field, None)
    
    def testInitWithParam(self):
        """Test the initialization if a fixed value of the digits is given."""
        digit = sg.Digit(2)
        
        self.assertEqual(digit.getFeasibleValues(), set([2]))
        self.assertEqual(digit.field, None)
    
    def testStringConversion(self):
        """Tests the stringification."""
        digit = sg.Digit()
        
        expected = '<(?,?)>'
        
        self.assertEqual(str(digit), expected)
    
    def testCard(self):
        """Tests the cardinality."""
        digit = sg.Digit()
        self.failUnlessEqual(digit.cardinality(), 10)
        
        digit = sg.Digit(1)
        self.failUnlessEqual(digit.cardinality(), 1)
    
    def testDone(self):
        """Tests whether isDone works."""
        digit = sg.Digit()
        digit.field = self.field
        
        self.failIf(digit.isDone())
        
        digit.force(set([1]))
        
        self.failUnless(digit.isDone())
        self.failUnlessEqual(digit.value(), 1)
    
    def testDoneWhenEmpty(self):
        """Tests whether isDone works."""
        digit = sg.Digit()
        digit.field = self.field
        
        self.failIf(digit.isDone())
        
        digit.force(set([]))
        
        self.failIf(digit.isDone())
    
    def testIsEmpty(self):
        """Tests whether isEmpty() works."""
        digit = sg.Digit()
        digit.field = self.field
        
        self.failIf(digit.isEmpty())
        
        digit.force(set([1]))
        self.failIf(digit.isEmpty())
        
        digit.force(set())
        self.assert_(digit.isEmpty())
    
    def testValidValues(self):
        """
        Test that Digit.getFeasibleValues() returns a copy of the valid 
        values.
        
        """
        digit = sg.Digit()
        
        self.assertEqual(digit.getFeasibleValues(), set(range(0,10)))
        #self.failIf(digit.getFeasibleValues() is digit.getFeasibleValues())
    
    def testRemove(self):
        """
        Test that Digit.invalidate() works correctly and can be rolled back.
        
        """
        digit = sg.Digit()
        digit.field = self.field
        
        self.assertEqual(digit.invalidate(0), True)
        self.assertEqual(digit.getFeasibleValues(), set(range(1,10)))
        
        self.assertEqual(digit.invalidate(0), False)
        self.assertEqual(digit.getFeasibleValues(), set(range(1,10)))
        
        undo.revert()
        
        self.assertEqual(digit.getFeasibleValues(), set(range(10)))
    

class DigitSplittingTest(test_utils.TestCase):
    def setUp(self):
        test_utils.TestCase.setUp(self)
        self.field = sg.Field(1,1)
    
    def testSplitAndForcingHalf(self):
        """Tests splitting and forcing"""
        
        sc.DIGIT_SPLIT = sc.DIGIT_SPLIT_HALF
        
        digit = self.field.digits[0][0]
        self.assert_(digit.isSplittable())
        
        self.assertEqual(digit.cardinality(), 10)
        
        parts = digit.split()
        self.assert_(len(parts) > 1)
        
        digit.force(parts[0])
        self.assert_(digit.cardinality() < 10)
        
        undo.revert()
        undo.startRevision()
                
        digit.force(parts[1])
        self.assert_(digit.cardinality() < 10)
    
    def testSplitAndForcingSingleton(self):
        """Tests splitting and forcing"""
        
        sc.DIGIT_SPLIT = sc.DIGIT_SPLIT_SINGLETON
        
        digit = self.field.digits[0][0]
        self.assert_(digit.isSplittable())
        
        self.assertEqual(digit.cardinality(), 10)
        
        parts = digit.split()
        self.assert_(len(parts) == 10)
        
        digit.force(parts[0])
        self.assert_(digit.cardinality() == 1)
        
        undo.revert()
        undo.startRevision()
                
        digit.force(parts[1])
        self.assert_(digit.cardinality() == 1)
    

class NumberTest(test_utils.TestCase):
    def testInit(self):
        """Test that the initialization works correctly."""
        number = sg.Number('A_w', 3)
        
        self.failUnlessEqual(number.name, 'A_w')
        
        self.failUnlessEqual(number.length, 3)
        
        self.failUnlessEqual(number.digits, [None, None, None])
        
        self.assertEqual(number.field, None)
    
    def testStringConversion(self):
        """Test the string conversion of numbers."""
        number = sg.Number('A_w', 3)
        
        self.assertEqual(str(number), 'A_w')
    
    def testIsEmptyTrue(self):
        """Test isEmpty() on a Number with one empty digit."""
        number = sg.Number('A_w', 3)
        number.digits = [sg.Digit(), sg.Digit(), sg.Digit()]
        number.digits[0].setFeasibleValues(set())
        
        self.assert_(number.isEmpty())
    
    def testIsEmptyFalse(self):
        """Test isEmpty() on a Number with no empty digit."""
        number = sg.Number('A_w', 3)
        number.digits = [sg.Digit(), sg.Digit(), sg.Digit()]
        
        self.failIf(number.isEmpty())
        self.assertEqual(number.cardinality(), 10*10*10)
    
    def testDone(self):
        """Tests whether isDone works."""
        number = sg.Number('A_w', 3)
        number.digits = [sg.Digit(), sg.Digit(), sg.Digit()]
        
        self.failIf(number.isDone())
        
        for digit in number.digits:
            digit.setFeasibleValues(set([1]))
        
        self.assert_(number.isDone())

    def testDoneEmptyDigits(self):
        """Tests whether isDone works."""
        number = sg.Number('A_w', 3)
        number.digits = [sg.Digit(), sg.Digit(), sg.Digit()]

        for digit in number.digits:
            digit.setFeasibleValues(set([]))
        
        self.failIf(number.isDone())
        self.assertSetEqual(number.getFeasibleValues(), set())
        self.failIf(number.isDone())

    def testPossibleValuesTest(self):
        """Tests getFeasibleValues()."""
        number = sg.Number('A_w', 3)
        
        # initialize digits
        number.digits[0] = sg.Digit()
        number.digits[0].setFeasibleValues(set([0, 1]))
        number.digits[1] = sg.Digit()   
        number.digits[1].setFeasibleValues(set([2, 3]))
        number.digits[2] = sg.Digit()   
        number.digits[2].setFeasibleValues(set([4, 5]))
        
        result = list(number.getFeasibleValues())
        expected = [124, 125, 134, 135, 24, 25, 34, 35]
        
        self.assertSetEqual(result, expected)
    
    def testPossibleValuesTest2(self):
        """Tests getFeasibleValues() again."""
        number = sg.Number('A_w', 3)
        
        # initialize digits
        number.digits[0] = sg.Digit()
        number.digits[0].setFeasibleValues(set([1, 2]))
        number.digits[1] = sg.Digit()
        number.digits[1].setFeasibleValues(set([1]))
        number.digits[2] = sg.Digit()
        number.digits[2].setFeasibleValues(set([1, 2]))
        
        result = number.getFeasibleValues()
        expected = [111,112,211,212]
        
        self.assertSetEqual(result, expected)
    
    def testMinimum(self):
        """Tests Number.minimum()."""
        number = sg.Number('A_w', 3)
        
        # initialize digits
        number.digits[0] = sg.Digit()
        number.digits[0].setFeasibleValues(set([0, 1]))
        number.digits[1] = sg.Digit()
        number.digits[1].setFeasibleValues(set([2, 3]))
        number.digits[2] = sg.Digit()
        number.digits[2].setFeasibleValues(set([4, 5]))
        
        result = number.minimum()
        expected = 24
        
        self.assertEqual(result, expected)
    
    def testMaximum(self):
        """Tests Number.minimum()."""
        number = sg.Number('A_w', 3)
        
        # initialize digits
        number.digits[0] = sg.Digit()
        number.digits[0].setFeasibleValues(set([0, 1]))
        number.digits[1] = sg.Digit()             
        number.digits[1].setFeasibleValues(set([2, 3]))
        number.digits[2] = sg.Digit()             
        number.digits[2].setFeasibleValues(set([4, 5]))
        
        result = number.maximum()
        expected = 135
        
        self.assertEqual(result, expected)
    
    def testMaximumOnExplicitTest(self):
        """Tests Number.minimum() on explicit test"""
        number = sg.Number('A_w', 3)
        self.field = sg.Field(1,1)
        number.field = self.field
        
        # initialize digits
        number.digits[0] = sg.Digit()
        number.digits[0].setFeasibleValues(set([0, 1]))
        number.digits[1] = sg.Digit()             
        number.digits[1].setFeasibleValues(set([2, 3]))
        number.digits[2] = sg.Digit()             
        number.digits[2].setFeasibleValues(set([4, 5]))
        
        number.force(set([123, 134]))
        
        result = number.maximum()
        expected = 134
        
        self.assertEqual(result, expected)
    
    def testCard(self):
        """Make sure that cardinality() works correctly."""
        number = sg.Number("A_w")
        self.assertEqual(sg.INFINITY, number.cardinality())
        
        number = sg.Number("A_w", 2)
        number.digits = [sg.Digit(), sg.Digit()]
        self.assertEqual(100, number.cardinality())
        
        number = sg.Number("A_w", 2)
        number.force(set([10,20,30]))
        self.assertEqual(3, number.cardinality())
    

class FieldNumberTest(test_utils.TestCase):
    def testInit(self):
        """Test that the initialization works correctly."""
        number = sg.FieldNumber('A_w', 1, 2, 3)
        
        self.failUnlessEqual(number.name, 'A_w')
        
        self.failUnlessEqual(number.pos, (1, 2))
        self.failUnlessEqual(number.length, 3)
        
        self.failUnlessEqual(number.digits, [None, None, None])
        
        self.assertEqual(number.field, None)

        # No digits have been assigned yet:
        self.failIf(number.hasDigits())
    

class DigitlessNumberTest(test_utils.TestCase):
    def setUp(self):
        test_utils.TestCase.setUp(self)
        self.field = sg.Field(2,2)
    
    def testInit(self):
        """Test IntermediateNumber initializations."""
        number = sg.Number('name')
        self.assertEqual(number.name, 'name')
        self.assertEqual(number._feasible_values, None)
        self.assertEqual(number.field, None)
    
    def testStringConversion(self):
        """Test that the string conversion works correctly."""
        number = sg.Number('name')
        self.assertEqual(str(number), "name")
    
    def testValue(self):
        """Test that the value() method works correctly."""
        number = sg.Number('name')
        self.assertNotEqual(number, None)
        self.assertEqual(number.value(), None)
        
        number.force(set([1]))
        self.assertEqual(number.value(), 1)
    
    def testIsDone(self):
        """Test the isDone() function of Number()"""
        number = sg.Number('X_0')
        self.failIf(number.isDone())
    
    def testIsEmpty(self):
        number = sg.Number('X_0')
        number.field = self.field
        
        self.assert_(not number.isEmpty())
        number.force(set())
        self.assert_(number.isEmpty())
        
        # must be rollbackable
        undo.revert()
        self.assert_(not number.isEmpty())
    
    def testValue(self):
        """Tests the value() function of Number()"""
        number = sg.Number('X', 2)
        number.field = self.field
        
        number.force(set([12]))
        self.assertEqual(number.value(), 12)
    
    def testSetValid(self):
        """Tests the force() function of Number()"""
        number = sg.Number('X', 2)
        number.field = self.field
        
        self.assert_(number.force(set([12])))
        self.assertEqual(number.value(), 12)
        self.failIf(number.force(set([12])))

    def testIntersect(self):
        number = sg.Number('X',2)
        number.field = self.field
        
        self.assert_(number.intersect(set([12,13,14])))
        self.assert_(number.intersect(set([12,13,15])))
        self.assertSetEqual(number.getFeasibleValues(), [12, 13])
        self.failIf(number.intersect(set([12,13,15])))
        self.assertSetEqual(number.getFeasibleValues(), [12, 13])

    def testRemoveSet(self):
        number = sg.Number('X',2)
        number.field = self.field
        
        self.failIf(number.invalidateSet(set([13,15])))
        self.assert_(number.force(set([12,13,14])))
        self.assert_(number.invalidateSet(set([13,15])))
        self.assertSetEqual(number.getFeasibleValues(), [12, 14])
        self.failIf(number.invalidateSet(set([13,15])))
        self.assertSetEqual(number.getFeasibleValues(), [12, 14])

    def testMarkInvalid(self):
        number = sg.Number('X',2)
        number.field = self.field
        
        self.failIf(number.invalidate(13))
        self.assert_(number.force(set([12,13,14])))
        self.assert_(number.invalidate(13))
        self.assertSetEqual(number.getFeasibleValues(), [12, 14])
        self.failIf(number.invalidate(13))
        self.failIf(number.invalidate(15))
        self.assertSetEqual(number.getFeasibleValues(), [12, 14])

    
    def testCard(self):
        """Tests the cardinality() function of Number()"""
        number = sg.Number('X', 2)
        
        d1 = sg.Digit()
        d2 = sg.Digit()
        number.digits = (d1,d2)
       
        self.assertEqual(number.cardinality(), 10 * 10)
        
        d1.setFeasibleValues(set([1]))
        d2.setFeasibleValues(set([2]))
        
        self.assertEqual(number.cardinality(), 1)
    


# Execute unit tests when called standalone.
if __name__ == "__main__":
    unittest.main()

# vim:sw=4:ts=4:sts=4:expandtab
